class UpdateWrapper:
    def __init__(self, update:dict, methods:object) -> None:
        self.update = update
        self.client = methods

    @property
    def object_guid(self) -> str:
        return self.update["chat_updates"][0].get("object_guid")
    
    @property
    def action(self) -> str:
        return self.update["chat_updates"][0].get("action")
    
    @property
    def chat_type(self) -> str:
        return self.update["chat_updates"][0].get("type")
    
    @property
    def count_unseen(self) -> int:
        return int(self.update["chat_updates"][0]["chat"].get("count_unseen", 0))
    
    @property
    def last_seen_peer_mid(self) -> str:
        return self.update["chat_updates"][0]["chat"].get("last_seen_peer_mid")
    
    @property
    def time_string(self) -> str:
        return self.update["chat_updates"][0]["chat"].get("time_string")
    
    @property
    def is_mine(self) -> bool:
        return self.update["chat_updates"][0]["chat"]["last_message"].get("is_mine")
    
    @property
    def time(self) -> str:
        return self.update["chat_updates"][0].get("object_guid")
    
    @property
    def status(self) -> str:
        return self.update["chat_updates"][0]["chat"].get("status")
    
    @property
    def last_message_id(self) -> str:
        return self.update["chat_updates"][0]["chat"].get("last_message_id")
    
    @property
    def message_id(self) -> str:
        return self.update["message_updates"][0].get("message_id")
    
    @property
    def reply_message_id(self) -> str:
        return self.update["message_updates"][0]["message"].get("reply_to_message_id")
    
    @property
    def text(self) -> str:
        return str(self.update["message_updates"][0]["message"].get("text"))
    
    @property
    def is_edited(self) -> bool:
        return self.update["message_updates"][0]["message"].get("is_edited")
    
    @property
    def message_type(self) -> str:
        if self.file_inline:
            return self.file_inline["type"]
        
        return self.update["message_updates"][0]["message"].get("type")
    
    @property
    def author_type(self) -> str:
        return self.update["message_updates"][0]["message"].get("author_type")
    
    @property
    def author_guid(self) -> str:
        return self.update["message_updates"][0]["message"].get("author_object_guid")
    
    @property
    def prev_message_id(self) -> str:
        return self.update["message_updates"][0].get("prev_message_id")
    
    @property
    def state(self) -> str:
        return self.update["message_updates"][0].get("state")
    
    @property
    def title(self) -> str:
        return self.update['show_notifications'][0].get('title')
    
    @property
    def author_title(self) -> str:
        return self.update['chat_updates'][0]['chat']['last_message'].get('author_title', self.title)
    
    @property
    def is_user(self) -> bool:
        return self.chat_type == "User"
    
    @property
    def is_group(self) -> bool:
        return self.chat_type == "Group"
    
    @property
    def is_forward(self) -> bool:
        return "forwarded_from" in self.update["message_updates"][0]["message"].keys()
    
    @property
    def forward_from(self) -> str:
        return self.update["message_updates"][0]["message"]["forwarded_from"].get("type_from") if self.is_forward else None
    
    @property
    def forward_object_guid(self) -> str:
        return self.update["message_updates"][0]["message"]["forwarded_from"].get("object_guid") if self.is_forward else None
    
    @property
    def forward_message_id(self) -> str:
        return self.update["message_updates"][0]["message"]["forwarded_from"].get("message_id") if self.is_forward else None

    @property
    def is_event(self) -> bool:
        return 'event_data' in self.update['message_updates'][0]['message'].keys()
    
    @property
    def event_type(self) -> str:
        return self.update['message_updates'][0]['message']['event_data'].get('type') if self.is_event else None
    
    @property
    def event_object_guid(self) -> str:
        return self.data['message_updates'][0]['message']['event_data']['performer_object'].get('object_guid') if self.is_event else None
    
    @property
    def pinned_message_id(self) -> str:
        return self.data['message_updates'][0]['message']['event_data'].get('pinned_message_id') if self.is_event else None
    
    @property
    def file_inline(self) -> dict:
        return self.update["message_updates"][0]["message"].get("file_inline")

    @property
    def has_link(self) -> bool:
        for link in ["http:/", "https:/", "www.", ".ir", ".com"]:
            if link in self.text.lower():
                return True
        return False
    
    def reply(self, text:str) -> dict:
        return self.client.send_text(object_guid=self.object_guid, text=text, message_id=self.message_id)
    
    def seen(self) -> dict:
        return self.client.seen_chats(seen_list={self.object_guid: self.message_id})
    
    def reaction(self, reaction:int) -> dict:
        return self.client.reaction_message(object_guid=self.object_guid, message_id=self.message_id, reaction=reaction)
    
    def delete(self, deleteForAll:bool=True) -> dict:
        return self.client.delete_messages(object_guid=self.object_guid, message_ids=[self.message_id], delete_for_all=deleteForAll)
    
    def pin(self) -> dict:
        return self.client.pin_message(object_guid=self.object_guid, message_id=self.message_id)
    
    def forward(self, to_object_guid:str) -> dict:
        return self.client.forward_messages(object_guid=self.object_guid, message_ids=[self.message_id], toObject_guid=to_object_guid)
    
    def ban(self) -> dict:
        return self.client.ban_member(object_guid=self.object_guid, member_guid=self.author_guid)
    
    def check_join(self, object_guid:str) -> bool:
        return self.client.check_join(object_guid=object_guid, user_guid=self.author_guid)
    
    def download(self, save:bool=False) -> dict:
        return self.client.download(save=save, file_inline=self.file_inline)


from .checkUpdate import checkUpdate
from .configs import Configs
from .exceptions import Exceptions
from .models import Models
from .sessions import Sessions
from .tools import Tools
from .updateWrapper import UpdateWrapper



from random import choices, randint
from time import time
from re import finditer, sub
from base64 import b64encode
from io import BytesIO
from tempfile import TemporaryFile
from mutagen import mp3, File
from filetype import guess
from os import system

class Tools:

    def randomTmpSession() -> str:
        return "".join(choices("abcdefghijklmnopqrstuvwxyz", k=32))
    
    def randomDeviceHash() -> str:
        return "".join(choices("0123456789", k=26))
    
    def randomRnd() -> str:
        return str(randint(-99999999, 99999999))
    
    def privateParse(private:str) -> str:
        if not private.startswith("-----BEGIN RSA PRIVATE KEY-----\n"):
            private = "-----BEGIN RSA PRIVATE KEY-----\n" + private
        
        if not private.endswith("\n-----END RSA PRIVATE KEY-----"):
            private += "\n-----END RSA PRIVATE KEY-----"

        return private

    def getState() -> int:
        return int(time()) - 150
    
    def phoneNumberParse(phoneNumber:str) -> str:
        if str(phoneNumber).startswith("0"):
            phoneNumber = phoneNumber[1:]

        elif str(phoneNumber).startswith("98"):
            phoneNumber = phoneNumber[2:]

        elif str(phoneNumber).startswith("+98"):
            phoneNumber = phoneNumber[3:]

        return phoneNumber
    
    def getChatTypeByGuid(objectGuid:str) -> str:
        for chatType in [("u0", "User"), ("g0", "Group"), ("c0", "Channel"), ("s0", "Service"), ("b0", "Bot")]:
            if objectGuid.startswith(chatType[0]): return chatType[1]
        return ""
    
    def getChatTypeByLink(link:str) -> str:
        if "rubika.ir/joing" in link: return "Group"
        elif "rubika.ir/joinc" in link: return "Channel"
    
    def checkMetadata(text):
        if text is None:
            return [], ""

        real_text = sub(r'``|\*\*|__|~~|--|@@|##|', '', text)
        metadata = []
        conflict = 0
        mentionObjectIndex = 0
        result = []

        patterns = {
            "Mono": r'\`\`([^``]*)\`\`',
            "Bold": r'\*\*([^**]*)\*\*',
            "Italic": r'\_\_([^__]*)\_\_',
            "Strike": r'\~\~([^~~]*)\~\~',
            "Underline": r'\-\-([^__]*)\-\-',
            "Mention": r'\@\@([^@@]*)\@\@',
            "Spoiler": r'\#\#([^##]*)\#\#',
        }

        for style, pattern in patterns.items():
            for match in finditer(pattern, text):
                metadata.append((match.start(), len(match.group(1)), style))

        metadata.sort()

        for start, length, style in metadata:
            if not style == "Mention":
                result.append({
                    'type': style,
                    'from_index': start - conflict,
                    'length': length,
                })
                conflict += 4
            else:
                mentionObjects = [i.group(1) for i in finditer(r'\@\(([^(]*)\)', text)]
                mentionType = Tools.getChatTypeByGuid(objectGuid=mentionObjects[mentionObjectIndex]) or 'Link'

                if mentionType == 'Link':
                    result.append(
                        {
                            'from_index': start - conflict,
                            'length': length,
                            'link': {
                                'hyperlink_data': {
                                    "url": mentionObjects[mentionObjectIndex]
                                },
                                'type': 'hyperlink',
                            },
                            'type': mentionType,
                        }
                    )
                else:
                    result.append(
                        {
                            'type': 'MentionText',
                            'from_index': start - conflict,
                            'length': length,
                            'mention_text_object_guid': mentionObjects[mentionObjectIndex],
                            'mention_text_object_type': mentionType
                        }
                    )
                real_text = real_text.replace(f'({mentionObjects[mentionObjectIndex]})', '')
                conflict += 6 + len(mentionObjects[mentionObjectIndex])
                mentionObjectIndex += 1
                
        return result, real_text
    
    def checkLink(url:str) -> dict:
        for i in ["http:/", "https://"]:
            if url.startswith(i): return True

    def getMimeFromByte(bytes:bytes) -> str:
        return (guess(bytes).extension or "pyrubi")
    
    def generateFileName(mime:str) -> str:
        return "Pyrubi Library {}.{}".format(randint(1, 1000), mime)
    
    def getImageSize(bytes:bytes) -> str:
        try:
            from PIL import Image
        except ImportError:
            system("pip install pillow")
            from PIL import Image

        width, height = Image.open(BytesIO(bytes)).size
        return width , height
    
    def getImageThumbnail(bytes:bytes) -> str:
        try:
            from PIL import Image
        except ImportError:
            system("pip install pillow")
            from PIL import Image
            
        image = Image.open(BytesIO(bytes))
        width, height = image.size
        if height > width:
            new_height = 40
            new_width  = round(new_height * width / height)
        else:
            new_width = 40
            new_height = round(new_width * height / width)
        image = image.resize((new_width, new_height), Image.LANCZOS)
        changed_image = BytesIO()
        image.save(changed_image, format='PNG')
        return b64encode(changed_image.getvalue()).decode('UTF-8')
    
    def getVideoData(bytes:bytes) -> list:
        try:
            import cv2
        except ImportError:
            system("pip install opencv-python")
            import cv2

        try:
            import numpy
        except ImportError:
            system("pip install numpy")
            import numpy

        with TemporaryFile(mode='wb+') as file:
            file.write(bytes)
            cap = cv2.VideoCapture(file.name)

            if not cap.isOpened():
                return None, None

            frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            frame_rate = int(cap.get(cv2.CAP_PROP_FPS))
            frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            video_duration = int(frame_count / frame_rate) * 1000

            ret, thumbnail_frame = cap.read()
            if not ret:
                return None, None
            
            encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), 1]
            _, buffer = cv2.imencode(".jpg", thumbnail_frame, encode_param)
            thumbnail_bytes = numpy.array(buffer).tobytes()

            cap.release()

            return b64encode(thumbnail_bytes).decode('UTF-8') , [frame_width, frame_height], video_duration
        
    def getVoiceDuration(bytes:bytes) -> int:
        file = BytesIO()
        file.write(bytes)
        file.seek(0)
        return mp3.MP3(file).info.length
    
    def getMusicArtist(bytes:bytes) -> str:
        try:
            audio = File(BytesIO(bytes), easy=True)

            if audio and 'artist' in audio:
                return audio['artist'][0]
            
            return "pyrubi"
        except Exception:
            return "pyrubi"


class Exceptions:
    
    class ClientException(Exception):
        pass

    class InvalidAuth(ClientException):

        def __init__(self, message:str="The Auth entered is invalid!"):
            super().__init__(message)

    class NotRegistered(ClientException):

        def __init__(self, message:str="Method input is not registered!"):
            super().__init__(message)

    class InvalidInput(ClientException):

        def __init__(self, message:str="Invalid method input!"):
            super().__init__(message)

    class TooRequests(ClientException):

        def __init__(self, message:str="Too much request! Your account has been suspended."):
            super().__init__(message)


from urllib3 import PoolManager
from json import loads

__version__ = "3.1.0"

def checkUpdate(appVersion:str=__version__) -> None:
    try:
        url:str = "https://aliganji1.github.io/api.pyrubi/checkUpdate.json"
        updateData = loads(PoolManager().request(method="GET", url=url).data.decode('UTF-8'))

        if updateData["version"] != appVersion:
            newUpdateText:str = '\u001b[33m' + f'\n{updateData["title"]}\nVersion: {updateData["version"]}\n\n{updateData["description"]}\n\nInstall: {updateData["update_command"]}\n' + '\u001b[37m'
            print(newUpdateText)
    except:pass


from json import dumps, loads
from tqdm import tqdm
from urllib3 import PoolManager
from websocket import create_connection, _exceptions as wsExceptions
from ..utils import Configs
from time import sleep
from threading import Thread, Event
from ..utils import Exceptions

class Network:

    def __init__(self, methods:object) -> None:
        self.methods = methods
        self.sessionData = methods.sessionData
        self.crypto = methods.crypto
        self.http = PoolManager(proxy_url=methods.proxy) if methods.proxy else PoolManager()
        self.event = Event()
        self.raise_info = {
            "INVALID_AUTH": Exceptions.InvalidAuth(),
            "NOT_REGISTERED": Exceptions.NotRegistered(),
            "INVALID_INPUT": Exceptions.InvalidInput(),
            "TOO_REQUESTS": Exceptions.TooRequests()
        }

    def getServer(self, type):
        if type == "api":
            from random import randint
            return f"https://messengerg2c{randint(2, 6)}.iranlms.ir"
        else:
            from random import choice
            return choice(list(loads(self.http.request("GET", "https://getdcmess.iranlms.ir/").data.decode())["data"]["socket"].values()))

    def request(self, method:str, input:dict={}, tmpSession:bool=False, apiVersion:int=6):
        url:str = self.getServer("api")

        data = {
            "api_version": apiVersion,
            "tmp_session" if tmpSession else "auth": self.crypto.auth if tmpSession else self.crypto.changeAuthType(self.sessionData["auth"]),
            "data_enc": self.crypto.encrypt(
                dumps({
                    "method": method,
                    "input": input,
                    "client": Configs.clients[self.methods.platform.lower()]
                })
            )
        }

        headers:dict = {
            "Accept-Encoding": "gzip",
            "Connection": "Keep-Alive",
            "User-Agent": "okhttp/3.12.1",
            "Referer": "https://web.rubika.ir/",
            "Content-Type": "application/json; charset=utf-8"
        }

        if self.sessionData:
            data["sign"] = self.crypto.makeSignFromData(data["data_enc"])

        while True:
            result = self.http.request(
                method="POST",
                url=url,
                headers=headers,
                body = dumps(data).encode(),
                timeout=self.methods.timeOut
            )

            result = loads(self.crypto.decrypt(loads(result.data.decode("UTF-8"))["data_enc"]))

            if result["status"] == "OK":
                if not self.sessionData: result["data"]["tmp_session"] = self.crypto.auth
                return result["data"]
            
            elif result["status"] in ["ERROR_GENERIC", "ERROR_ACTION"]:
                raise self.raise_info[result["status_det"]]
            
            continue

    def handShake(self):
        global ws

        def sendData(mode:int=0):
            if mode == 0:
                ws.send(dumps(
                    {
                        "api_version": "5",
                        "auth": self.sessionData["auth"],
                        "method": "handShake"
                    }
                ))

            if mode == 1:
                ws.send("{}")

        def keepAlive():
            from ..utils import Exceptions

            while True:
                try:
                    sendData(1)
                    self.methods.getChatsUpdates()
                    sleep(60)
                except Exceptions.NotRegistered or Exceptions.InvalidAuth:
                    print("Invalid session data! Please enter a valid data of your account.")
                    exit()
                except Exceptions.TooRequests:
                    break
                except:
                    continue

        print("Connecting to the web socket...")
        wsServer:str = self.getServer("socket")

        while True:
            try:
                ws = create_connection(url=wsServer)

                sendData()
                if loads(ws.recv()).get("status") == "OK":
                    print("Connected.")

                    Thread(target=keepAlive, daemon=True).start()

                    while True:
                        recv:dict = loads(ws.recv())

                        if recv.get("type") == "messenger":
                            yield loads(self.crypto.decrypt(recv.get("data_enc")))

            except wsExceptions.WebSocketConnectionClosedException or wsExceptions.WebSocketAddressException:
                print("The websocket connection was disconnected! Reconnecting...")
                continue

    def upload(self, file:str and bytes, fileName:str=None, chunkSize:int=131072):
        from ..utils import Tools

        if isinstance(file, str):
            if Tools.checkLink(url=file):
                file:bytes = self.http.request("GET", file).data
                mime:str = Tools.getMimeFromByte(file)
                fileName = fileName or Tools.generateFileName(mime=mime)
            else:
                fileName = fileName or file
                mime = file.split(".")[-1]
                file = open(file, "rb").read()

        elif not isinstance(file, bytes):
            raise FileNotFoundError("Enter a valid path or url or bytes of file.")
        else:
            fileName = Tools.generateName()
            mime = Tools.getMimeFromByte(file)

        def send_chunk(data, maxAttempts=2):
            for attempt in range(maxAttempts):
                try:
                    response = self.http.request(
                        "POST",
                        url=requestSendFileData["upload_url"],
                        headers=header,
                        body=data
                    )
                    return loads(response.data.decode("UTF-8"))
                except Exception:
                    print(f"\nError uploading file! (Attempt {attempt + 1}/{maxAttempts})")
            
            print("\nFailed to upload the file!")

        requestSendFileData:dict = self.methods.requestSendFile(
            fileName = fileName,
            mime = mime,
            size = len(file)
        )

        header = {
            "auth": self.sessionData["auth"],
            "access-hash-send": requestSendFileData["access_hash_send"],
            "file-id": requestSendFileData["id"],
        }

        totalParts = (len(file) + chunkSize - 1) // chunkSize

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Uploading {fileName}",
                total=len(file),
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        for partNumber in range(1, totalParts + 1):
            startIdx = (partNumber - 1) * chunkSize
            endIdx = min(startIdx + chunkSize, len(file))
            header["chunk-size"] = str(endIdx - startIdx)
            header["part-number"] = str(partNumber)
            header["total-part"] = str(totalParts)
            data = file[startIdx:endIdx]
            hashFileReceive = send_chunk(data)
            if self.methods.showProgressBar:
                processBar.update(len(data))
            if not hashFileReceive: return
            if partNumber == totalParts:
                if not hashFileReceive["data"]: return
                requestSendFileData["file"] = file
                requestSendFileData["access_hash_rec"] = hashFileReceive["data"]["access_hash_rec"]
                requestSendFileData["file_name"] = fileName
                requestSendFileData["mime"] = mime
                requestSendFileData["size"] = len(file)
                return requestSendFileData
            
    def download(self, accessHashRec:str, fileId:str, dcId:str, size:int, fileName:str, chunkSize:int=131072):

        header:dict = {
            "Connection": "Keep-Alive",
            "Content-Type": "application/json",
            "User-Agent": "okhttp/3.12.1",
            "client-app-name": "Main",
            "client-app-version": "3.5.4",
            "app.rbmain.a": "app.rbmain.a",
            "client-platform": "Android",
            "auth": self.sessionData["auth"],
            "access-hash-rec": accessHashRec,
            "file-id": fileId,
            "start-index": "0"
        }

        data:bytes = b""

        def downloadChunk(maxAttempts=2) -> bytes:
            for attempt in range(maxAttempts):
                try:
                    return self.http.request(
                        "GET",
                        url=f"https://messenger{dcId}.iranlms.ir/GetFile.ashx",
                        headers=header,
                    ).data
                except Exception:
                    print(f"\nError downloading file! (Attempt {attempt + 1}/{maxAttempts})")
            
            print("\nFailed to download the file!")

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Downloading {fileName}",
                total=size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        while True:
            lastIndex = int(header["start-index"]) + chunkSize - 1 if int(header["start-index"]) + chunkSize < size else size - 1
            header["last-index"] = str(lastIndex)

            downloadedData = downloadChunk()
            if downloadedData:
                data += downloadedData
                if self.methods.showProgressBar:
                    processBar.update(len(downloadedData))

            if len(data) >= size:
                return data

            header["start-index"] = lastIndex + 1